import csv
import logging
from sys import stderr
import os

import click

from ontoma.constants import RESULT_FIELDS, EFO_DEFAULT_VERSION
from ontoma.interface import OnToma

logger = logging.getLogger()


def get_version() -> str:
    """
    Reads the VERSION file in the same directory as the script and returns its contents.

    Returns:
        The version number of the script.
    """
    script_path = os.path.dirname(__file__)
    return open(f"{script_path}/../VERSION").read().strip()


@click.command()
@click.option(
    "--infile",
    type=click.File("r"),
    default="/dev/stdin",
    help="File to read the input from, one per line. The type of input is determined by --input-type.",
)
@click.option(
    "--outfile",
    type=click.File("w"),
    default="/dev/stdout",
    help="File to output the mappings to. The format is two column TSV file. First column is your query, and the "
    "second column contains the mappings. When no mappings were found, no lines will be output. When multiple "
    "mappings are found, multiple lines will be output, with one mapping per line.",
)
@click.option(
    "--input-type",
    type=click.Choice(["string", "ontology"]),
    default="string",
    help="Can be one of two values. When set to “string” (default), input values are treated as strings, e.g. "
    "“Diabetes”. When set to “ontology”, input values are treated as ontology identifiers, e.g. “MONDO_0005015”.",
)
@click.option(
    "--cache-dir",
    type=str,
    default=None,
    help="A directory to store EFO cache. Specifying it is not required, but will speed up subsequent OnToma runs.",
)
@click.option(
    "--columns",
    type=str,
    default="query,id_ot_schema",
    help=f'Which columns to output, comma separated. The available options are: {",".join(RESULT_FIELDS)}',
)
@click.option(
    "--efo-release",
    type=str,
    default=EFO_DEFAULT_VERSION,
    help=f"EFO release to use. This must be be either “latest”, or match the specific tag name in their GitHub "
    f"releases, for example v3.31.0. By default, {EFO_DEFAULT_VERSION!r} is used.",
)
@click.option("--version", help="Print version number and exit.", is_flag=True)
@click.option(
    "--log-level",
    type=click.Choice(["INFO", "DEBUG", "WARN", "ERROR"]),
    default="INFO",
    help="Log verbosity level.",
)
def ontoma(
    infile, outfile, input_type, cache_dir, columns, efo_release, version, log_level
):
    """Maps ontology identifiers and strings to EFO, the ontology used by the Open Targets Platform."""
    # Initialize logger:
    logging.basicConfig(
        level=getattr(logging, log_level),
        format="%(asctime)s %(levelname)s %(module)s - %(funcName)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if version:
        stderr.write(f"OnToma version {get_version()}.\n")
        return

    if infile.name == "/dev/stdin":
        logger.warning(
            "Reading input from STDIN. If this is not what you wanted, re-run with --help to see usage."
        )

    logger.info("Initialising OnToma main interface.")
    otmap = OnToma(cache_dir=cache_dir, efo_release=efo_release)
    columns = columns.split(",")
    efo_writer = csv.DictWriter(outfile, columns, delimiter="\t")
    efo_writer.writeheader()

    logger.info(f"Treating the input as (type = {input_type}) and mapping to EFO.")
    mapped = 0
    failed = 0
    for line in infile:
        query = line.rstrip()
        results = otmap.find_term(query, code=input_type == "ontology")
        for result in results:
            efo_writer.writerow({c: getattr(result, c) for c in columns})
        if results:
            mapped += 1
        else:
            failed += 1

    logger.info(
        f"Processed {mapped + failed} inputs. Of them, found at least one EFO hit for {mapped}, and failed to find any hits for {failed}."
    )



"""Wrapper for the OxO API."""

import logging

import requests

__all__ = [
    'OxoClient'
]

OXO = 'https://www.ebi.ac.uk/spot/oxo/api'

logger = logging.getLogger(__name__)


class OxoClient:
    """OxO wrapper class."""

    def __init__(self, base_url=OXO.rstrip('/')):
        self._baseapi = base_url
        self._searchapi = base_url + '/search'

    @staticmethod
    def _pages(method, url, *args, **kwargs):
        """Generator expression to paginate through OxO results."""
        r = method(url, *args, **kwargs)
        r.raise_for_status()
        resp = r.json()
        yield resp
        if 'next' not in resp['_links']:
            return
        while 'next' in resp['_links']:
            logger.debug(f"Found next page at {resp['_links']['next']['href']}")
            r = method(resp['_links']['next']['href'], *args, **kwargs)
            r.raise_for_status()
            resp = r.json()
            yield resp

    def search(self, ids=None, mapping_target='EFO', distance=1):
        """Query OxO with given parameters and yield the found mappings one by one."""
        payload = {
            'ids': ids,
            'mappingTarget': mapping_target,
            'distance': str(distance)
        }
        logger.debug(f'Querying OxO /search with payload: {payload}')
        for page in self._pages(requests.post, self._searchapi, data=payload):
            logger.debug('/search - Returned {} mappings'.format(len(page['_embedded']['searchResults'])))
            for mapping in page['_embedded']['searchResults']:
                yield mapping



__all__ = ['OnToma']

import logging

from ontoma.interface import OnToma

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(levelname)-8s - %(name)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)



"""Constants for URLs and other values."""

# Manual mapping databases.
URLS = {
    'MANUAL_XREF': 'https://raw.githubusercontent.com/opentargets/curation/master/mappings/disease/manual_xref.tsv',
    'MANUAL_STRING': 'https://raw.githubusercontent.com/opentargets/curation/master/mappings/disease/manual_string.tsv',
}

# List of fields to available for result output.
RESULT_FIELDS = ('query', 'id_normalised', 'id_ot_schema', 'id_full_uri', 'label')

EFO_DEFAULT_VERSION = 'latest'



"""Class to handle EFO ontology related functions."""

import logging
import os
import urllib.request

import pandas as pd
import pronto
import requests
from retry import retry
from typing import Dict

from ontoma import ontology

logger = logging.getLogger()


class EFOHandler:
    """A collection of functions to manage EFO data retrieving, version control and caching."""

    EFO_RELEASE_API_TEMPLATE = "https://api.github.com/repos/EBISPOT/efo/releases/{}"

    # List of cached datasets:
    DATASETS = {"terms", "xrefs", "synonyms"}

    # The OWL filename is fixed and won't change by versions:
    OWL_FILENAME = "efo_otar_slim.owl"

    def __init__(self, efo_release: str, cache_dir: str) -> None:
        """By providing and EFO version and cache dir, the class tries to load cached data or create new cache.

        Args:
            efo_release (str): EFO version. Must be a valid release tag.
            cache_dir (str): Folder into which the cache dataset will be saved.
        """
        # Get EFO release related meta data:
        efo_release_info = self.__get_efo_github_data(efo_release)
        self.efo_version = efo_release_info["tag_name"]
        self.assets = efo_release_info["assets"]
        self.cache_dir = cache_dir

        # If cache dir doesn't exist, create it:
        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)
            logger.info(f"Created EFO cache directory {cache_dir}.")

        # Try to load all the files, if any of the file is missing, rebuilding cache:
        for dataset_name in self.DATASETS:
            try:
                setattr(
                    self,
                    dataset_name,
                    self.__load_cache(dataset_name),
                )
            except FileNotFoundError:
                logger.info(
                    f"{dataset_name} could not be build from cache. Re-building cache."
                )
                self.__generate_cache()
                break

    def __save_cache(self, dataset_name: str, dataset: pd.DataFrame) -> None:
        # Generate file name:
        file_name = os.path.join(
            self.cache_dir, f"{dataset_name}.{self.efo_version}.tsv"
        )
        # Save pandas datafarme:
        logger.info(f"Saving {dataset_name} to {file_name}")
        dataset.to_csv(file_name, sep="\t", index=False)

    def __load_cache(self, dataset_name: str) -> pd.DataFrame:
        file_name = os.path.join(
            self.cache_dir, f"{dataset_name}.{self.efo_version}.tsv"
        )
        logger.info(f"Loading {dataset_name} from cache: {file_name}")
        return pd.read_csv(file_name, sep="\t")

    def __generate_cache(self) -> None:
        otar_slim_assets = [a for a in self.assets if a["name"] == self.OWL_FILENAME]
        if len(otar_slim_assets) == 0:
            raise AssertionError(
                f"EFO release {self.efo_version!r} on GitHub does not contain the file {self.OWL_FILENAME!r}."
            )
        if len(otar_slim_assets) > 1:
            raise AssertionError(
                f"EFO release {self.efo_version!r} contains multiple files named {self.OWL_FILENAME!r}."
            )
        otar_slim_url = otar_slim_assets[0]["browser_download_url"]

        # Retrieve OWL file to cache dir:
        logging.info(f"Fetching the OWL file ({self.OWL_FILENAME}).")
        urllib.request.urlretrieve(
            otar_slim_url, os.path.join(self.cache_dir, self.OWL_FILENAME)
        )

        # Processing owl file:
        logging.info("Processing the OWL file.")
        terms_dataset, xrefs_dataset, synonyms_dataset = [], [], []

        # Reading ontology:
        for term in pronto.Ontology(
            os.path.join(self.cache_dir, self.OWL_FILENAME)
        ).terms():
            normalised_term_id = ontology.normalise_ontology_identifier(term.id)
            terms_dataset.append(
                [
                    normalised_term_id,
                    term.name.lower() if term.name else "",
                    term.obsolete,
                ]
            )
            # Extracting cross references:
            for xref in term.xrefs:
                normalised_xref_id = ontology.normalise_ontology_identifier(xref.id)
                if normalised_xref_id:
                    xrefs_dataset.append(
                        [
                            normalised_xref_id,
                            normalised_term_id,
                        ]
                    )
            # Extracting synonyms:
            for synonym in term.synonyms:
                if synonym.description and synonym.scope == "EXACT":
                    synonyms_dataset.append(
                        [
                            synonym.description.lower(),
                            normalised_term_id,
                        ]
                    )

        # Store datasets:
        self.terms = pd.DataFrame(
            terms_dataset, columns=("normalised_id", "normalised_label", "is_obsolete")
        )
        self.xrefs = pd.DataFrame(
            xrefs_dataset,
            columns=(
                "normalised_xref_id",
                "normalised_id",
            ),
        )
        self.synonyms = pd.DataFrame(
            synonyms_dataset, columns=("normalised_synonym", "normalised_id")
        )

        # Looping through all datasets and save cache:
        logging.info("Save cache.")
        for dataset_name in self.DATASETS:
            self.__save_cache(dataset_name, getattr(self, dataset_name))

    def get_terms(self) -> pd.DataFrame:
        """This function returns the EFO term list as a pandas DataFrame."""
        return self.terms

    def get_xrefs(self) -> pd.DataFrame:
        """This function returns the EFO term cross references as a pandas DataFrame."""
        return self.xrefs

    def get_synonyms(self) -> pd.DataFrame:
        """This function returns the EFO term synonyms as pandas DataFrame."""
        return self.synonyms

    @retry(logger=logger, tries=5, delay=3, backoff=1.5, jitter=(1, 3))
    def __get_efo_github_data(self, efo_release: str) -> Dict[str, str]:
        """Queries GitHub API to fetch the latest EFO release."""
        if efo_release == "latest":
            url = self.EFO_RELEASE_API_TEMPLATE.format(efo_release)
        else:
            url = self.EFO_RELEASE_API_TEMPLATE.format(f"tags/{efo_release}")
        response = requests.get(url)
        response.raise_for_status()  # In case of HTTP errors, this will be caught by the @retry decorator.
        return response.json()

    def get_efo_version(self) -> str:
        """Return the cached EFO version."""
        return self.efo_version



"""Functions that download mapping flat files from github repo or other repositories."""

__all__ = [
    'get_manual_xrefs',
    'get_manual_string_mappings',
]

import logging

import pandas as pd

from ontoma.ontology import normalise_ontology_identifier

logger = logging.getLogger(__name__)


def get_manual_xrefs(url: str) -> pd.DataFrame:
    """Download the manual cross-reference list and convert to a Pandas dataframe."""
    logger.debug(f'Requesting manual ontology-to-EFO mappings from URL {url}')
    df = pd.read_csv(url, sep='\t')
    df.normalised_xref_id = df.normalised_xref_id.apply(normalise_ontology_identifier)
    df.normalised_id = df.normalised_id.apply(normalise_ontology_identifier)
    return df


def get_manual_string_mappings(url: str) -> pd.DataFrame:
    """Download the manual string-to-ontology mapping list and convert to a Pandas dataframe."""
    logger.debug(f'Requesting manual string-to-EFO mappings from URL {url}')
    df = pd.read_csv(url, sep='\t')
    df['normalised_label'] = df.PROPERTY_VALUE.apply(str.lower)
    df['normalised_id'] = df.SEMANTIC_TAG.apply(normalise_ontology_identifier)
    return df[['normalised_label', 'normalised_id']]



import os
import sys
from setuptools import setup, find_packages
from setuptools.command.install import install

with open('VERSION') as version_file:
    VERSION = version_file.read().strip()

with open('README.md') as f:
    readme = f.read()


class VerifyVersionCommand(install):
    description = 'Verify that the git tag matches the VERSION file.'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')
        if tag != VERSION:
            sys.exit(f'Git tag {tag} does not match the version of this app {VERSION}')


setup(
    name='ontoma',
    version=VERSION,
    description='Ontology mapping for Open Targets',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Open Targets data team',
    author_email='data@opentargets.org',
    url='https://github.com/opentargets/OnToma',
    license='Apache License, Version 2.0',
    packages=find_packages(exclude=('tests', 'docs')),
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords='opentargets ontology efo mapper',
    install_requires=[
        'click',
        'pandas',
        'pronto',
        'requests',
        'retry2',
    ],
    entry_points='''
        [console_scripts]
        ontoma=ontoma.cli:ontoma
    ''',
    python_requires='>=3.7',
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)
